import path from 'node:path'
import { readFileSync, writeFileSync } from 'node:fs'
import { describe, it, expect, beforeEach, afterEach } from '@jest/globals'
import { createServer } from 'vite'
import vitePluginOxExternals from '@open-xchange/vite-plugin-ox-externals'

import vitePluginOxManifests from '../../dist/index'
import { waitForReload } from '../../dist/util'
import { getPort, getDirname } from '../util'

const __dirname = getDirname(import.meta.url)
const files = ['register.js', 'settings.js']
const PORT = getPort()

describe('HMR', function () {
  let server, data

  beforeEach(function () {
    data = files.map(f => readFileSync(path.join(__dirname, f)))
  })

  afterEach(async function () {
    data.forEach((d, i) => writeFileSync(path.join(__dirname, files[i]), d))
    data = null
    await server?.close()
    server = null
  })

  it('adds timestamps to manifest files', async function () {
    server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginOxManifests({

      }), vitePluginOxExternals({
        prefix: '$'
      })]
    })
    await server.listen(PORT)

    // fetch this file to start watcher
    await fetch(`http://localhost:${PORT}/register.js`)

    const manifests = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifests)).toHaveLength(2)
    expect(manifests[1]).not.toContain('raw')

    writeFileSync(path.join(__dirname, 'register.js'), 'console.log("Changed")')

    await waitForReload(server)

    const manifestsWithHMR = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifestsWithHMR)).toHaveLength(2)
    expect(manifestsWithHMR[1].namespace).toEqual('test')
    expect(manifestsWithHMR[1].raw).toMatch(/\/register\.js\?t=\d+/)
  })

  it('adds timestamps to settings manifest files', async function () {
    server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginOxManifests({

      }), vitePluginOxExternals({
        prefix: '$'
      })]
    })
    await server.listen(PORT)

    // fetch this file to start watcher
    await fetch(`http://localhost:${PORT}/settings.js`)

    const manifests = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifests)).toHaveLength(2)
    expect(manifests[0]).not.toContain('raw')

    writeFileSync(path.join(__dirname, 'settings.js'), `/* a comment */${data[1].toString()}`)

    await waitForReload(server)

    const manifestsWithHMR = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifestsWithHMR)).toHaveLength(2)
    expect(manifestsWithHMR[0].namespace).toEqual('settings')
    expect(manifestsWithHMR[0].raw).toMatch(/\/settings\.js\?t=\d+/)
  })

  it('includes base-path in manifests', async function () {
    server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite',
      plugins: [vitePluginOxManifests({

      }), vitePluginOxExternals({
        prefix: '$'
      })]
    })
    await server.listen(PORT)

    // fetch this file to start watcher
    await fetch(`http://localhost:${PORT}/appsuite/register.js`)
    await fetch(`http://localhost:${PORT}/appsuite/settings.js`)

    const manifests = await fetch(`http://localhost:${PORT}/appsuite/manifests`).then(r => r.json())
    expect(Object.keys(manifests)).toHaveLength(2)
    expect(manifests[0]).not.toContain('raw')
    expect(manifests[1]).not.toContain('raw')

    writeFileSync(path.join(__dirname, 'register.js'), 'console.log("Changed")')
    await waitForReload(server)

    writeFileSync(path.join(__dirname, 'settings.js'), `/* a comment */${data[1].toString()}`)
    await waitForReload(server)

    const manifestsWithHMR = await fetch(`http://localhost:${PORT}/appsuite/manifests`).then(r => r.json())
    expect(Object.keys(manifestsWithHMR)).toHaveLength(2)
    expect(manifestsWithHMR[0].namespace).toEqual('settings')
    expect(manifestsWithHMR[0].raw).toMatch(/\/appsuite\/settings\.js\?t=\d+/)
    expect(manifestsWithHMR[1].namespace).toEqual('test')
    expect(manifestsWithHMR[1].raw).toMatch(/\/appsuite\/register\.js\?t=\d+/)
  })
})
